// MQOCOMM.CPP program

//*********************************************************************************************
//*                                                                                           *
//*                            MQOCOMM.CPP program                                            *
//*                            ====================                                           *
//*                                                                                           *
//*   This program will issue a Command to the Queue Manager, displaying the result.          *
//*        The Command Server must be running to perform the command.                         *
//*                                                                                           *
//*                                                                                           *
//*   The parameters are  /Maaaaaa    -> the name of the source Queue Manager                 *
//*                                                  default -> blanks (default QM)           *
//*                                                                                           *
//*                       /Qaaaaaa    -> the command input Queue to use                       *
//*                                                  default -> SYSTEM.ADMIN.COMMAND.QUEUE    *
//*                                                                                           *
//*                       /Paaaaaa    -> the model for the ReplyToQ to use                    *
//*                                                  default -> SYSTEM.MQSC.REPLY.QUEUE       *
//*                                                                                           *
//*                       /Raaaaaa    -> the name of the ReplyToQ to use                      *
//*                                                  default -> MQOCOMM.*                     *
//*                                                                                           *
//*                       /Naaaaaa    -> the name of the ReplyToQ Manager to use              *
//*                                                  default -> /M's Queue Manager            *
//*                                                                                           *
//*                       /Uaaaaaa    -> the USERID of the issuer of the Message to the       *
//*                                          Comand Server. This defaults to X'00's, and      *
//*                                          if used to systems which support Security must   *
//*                                          be valid and authorised etc.                     *
//*                                                                                           *
//*                       /C+         -> Model ReplyToQ used (/R), create the ReplyToQ (/S)   *
//*                                            (default)                                      *
//*                       /C-         -> Model ReplyToQ unused, access /S ReplyToQ            *
//*                                                                                           *
//*                       /A=         -> Send Command & Receive Reply (default)               *
//*                       /A+         -> Send Command                                         *
//*                       /A-         ->                Receive Reply                         *
//*                                                                                           *
//*                                                                                           *
//*                                                                                           *
//*                       /Tnnnnn     -> Timeout for ReplyToQ MQGet                           *
//*                                                  default -> 5 seconds                     *
//*                                                                                           *
//*                       aaaaaaaaa   -> the command to issue (must come after / parms)       *
//*                                                  default -> blanks                        *
//*                                                                                           *
//*                       /V+         -> Write Diagnotic Messages                             *
//*                       /V-         -> Don't display any messages (default)                 *
//*                                                                                           *
//*   Example : MQOCOMM  /MRAH1    QUERY QUEUE(A) ALL        will issue the given command to  *
//*                                                          Queue Manager RAH1.              *
//*                                                                                           *
//*  This utility will work in one of two ways, keyed upon the /A parameter:                  *
//*                                                                                           *
//*       /A= is the normal way of operation. It sends a command to the command               *
//*           Server Queue, and waits for a response to appear in the ReplyToQ.               *
//*                                                                                           *
//*           The ReplyToQ can be predefined (use /C-, with /R specifing the name)            *
//*               or it can be generated for the command ( /C+ ). If the ReplyToQ             *
//*               is generated, then its attributes are taken from the Model Q                *
//*               given in /P. The NAME of this ReplyToQ is generated from the                *
//*               /R parameter. If /R ends in a *, then a random unique name is               *
//*               generated by MQ; however, if it does not end in the *, the                  *
//*               Q must not exist.                                                           *
//*                                                                                           *
//*       /A+ and /A- seperate the two functions of /A=, into Send (/A+) and                  *
//*           Receive (/A-) operations. This allows seperation of the Sending of              *
//*           the command and the receipt of its response.                                    *
//*                                                                                           *
//*       /A+ sends the command to the Command Queue named in /Q. The ReplyToQ                *
//*           may already exist ( /C- ) and is named by /R. Alternatively. the                *
//*           ReplyToQ can be generated from the Model given in /P ; however, the             *
//*           Model must be PERMDYN in order for the ReplyToQ to be kept for                  *
//*           subsequent reading (MQOCOMM does not check this). If the pattern                *
//*           for the name of the to be created ReplyToQ ends with a * (indicating            *
//*           the generation of a guarenteed unique name), then /V+ must be                   *
//*           used to obtain the generated name. If the name given for the ReplyToQ           *
//*           in /R does not end in a *, then the named Queue must not exist (so              *
//*           that the Model can create it).                                                  *
//*                                                                                           *
//*       /A- reads a Queue named in /R to extract the responses to a prior                   *
//*           command which put replies into this Queue. In this case, /C+ is                 *
//*           forced to /C- to show that the name of the Queue to read must be                *
//*           known.                                                                          *
//*                                                                                           *
//*   The /A /C /R and /P parameters thus interact as follows:                                *
//*                                                                                           *
//*   /A=   /C+    /R ends in *            /P used; PERMDYN or TEMPDYN   (default)            *
//*                /R does not end in *,   /P used; PERMDYN or TEMPDYN                        *
//*                    and does not exist                                                     *
//*                                                                                           *
//*   /A=   /C-    /R has to exist         /P not used                                        *
//*                                                                                           *
//*   /A+   /C+    /R should not end       /P used; got to be PERMDYN                         *
//*                    in *, and does not                                                     *
//*                    exist                                                                  *
//*                                                                                           *
//*   /A+   /C-    /R has to exist         /P not used                                        *
//*                                                                                           *
//*   /A-   /C+    forced to /C-                                                              *
//*                                                                                           *
//*   /A-   /C-    /R has to exist         /P not used                                        *
//*                                                                                           *
//*                                                                                           *
//*   The displayed output (sent to COUT) will be prefixed by ! if this utility               *
//*       has generated the line (either in reponse to /V+ being used, or things              *
//*       that always come out). Lines NOT starting with ! have come from the                 *
//*       Command Server (via the ReplyToQ); however, only those responses which              *
//*       are string-types (see the PCF manual for details on this) are                       *
//*       displayed. Use /V+ to display all the integer-type messages as well.                *
//*                                                                                           *
//*                                                                                           *
//*   If operating in seperate Send / Receive mode, then the Receive action will only         *
//*      extract a SINGLE message set at a time. This is deliberately done so as              *
//*      not to confuse things. You are at liberty to alter the code to see if                *
//*      there is more to extract if you want so to do.                                       *
//*                                                                                           *
//*        When the FOR loop exit is taken ( with type L ), I suggest you do an               *
//*        rmq.inq on the number of messages still left in the ReplyToQ, and if               *
//*        non-zero is returned, go around the FOR loop again.                                *
//*                                                                                           *
//*        However, I do not approve of this action, and so have not provided an              *
//*        option for it!                                                                     *
//*                                                                                           *
//*                                                                                           *
//*   If you are sending the command to ANOTHER Queue Manager (by using /Q which              *
//*      points to a Transmission Queue), you can use /N to specify where the                 *
//*      ReplyToQ lives; this enables the usual routing operations to be                      *
//*      performed for the ReplyToQ (ie: QM -> Transmission Q). If /N is omitted,             *
//*      then the Queue Manager named in /M is used. This facility to use different           *
//*      Queue Managers is NOT available for /A= (Send & Receive) operations, and             *
//*      only has relevance for /A+ (Send only) operation (as /A- will use the QM             *
//*      named in /M). Additionally, if /M and /N disagree, /C- (ReplyToQ already             *
//*      exists) is forced, as it is not possible to create a ReplyToQ in these               *
//*      circumstances.                                                                       *
//*                                                                                           *
//*      You should also be aware that this utility has only been tested under                *
//*      the OS/2 version of MQ. However, it should work for the UNIX flavours                *
//*      (the IBM ones, not ezBridge). It WILL not work if the QM that receives               *
//*      the command is on MVS - as at the time of coding the MVS version of MQ               *
//*      does NOT support either PCF flows and so the ESCAPE PCF operation will               *
//*      not work.  Similiarly, the /U parameter which supplies a Userid for the              *
//*      Command server to authorise the command has no meaning in the OS/2                   *
//*      environment (as Security is not provided).                                           *
//*                                                                                           *
//*                                                                                           *
//*   If NOTHING comes back from this utility, then check that the Command server             *
//*      is running (AMQPCSEA). If it is, then bump up the /T timeout setting to              *
//*      something bigger, and repeat the command with /V+; then look at the Return           *
//*      Codes and Diagnostic messages to see what caused the failure.                        *
//*                                                                                           *
//*                                                                                           *
//*   Return Codes :    0    -> Operation succeeded                                           *
//*                     >0   -> Reason Code for a failure                                     *
//*                                                                                           *
//*                                                                                           *
//*   Copyright : This program is Copyrighted by IBM UK Labs Limited, 1995.                   *
//*                                                                                           *
//*                                                                                           *
//*   Support   : This program is not supported in any way by IBM. It is distributed          *
//*               only to show techniques for Message Queueing. It may be freely              *
//*               modified and adapted by your installation.                                  *
//*                                                                                           *
//*   Author    : Robert Harris,                                                              *
//*               TS/OO Technology,                                                           *
//*               IBM Hursley Park,                                                           *
//*               England.                                                                    *
//*                                                                                           *
//*********************************************************************************************

//
// System Includes
//

#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <time.h>

//
// MQ Includes
//

#include <cmqc.h>
#include <cmqcfc.h>
#include <cmqxc.h>

//
// OS2 Includes
//

#define  INCL_DOSMEMMGR
#include <os2.h>

//
// Definitions of datatypes
//

#define FULLWORD     long
#define HALFWORD     int
#define MAXSFULLWORD (68719476735)
#define MAXUFULLWORD (4294967295)
#define MAXSHALFWORD (4095)
#define MAXUHALFWORD (65535)
#define MAXFILENAME  (255)
#define MAXCOMMLEN   (2000)

//
// The parms class contains the unravelling of the
//     supplied parameters
//

class parms
{
   private:

       char        cqm ??(MQ_Q_MGR_NAME_LENGTH+1??) ; // QM    name - Command
       char        cq  ??(MQ_Q_NAME_LENGTH+1    ??) ; // Q     name - Command
       char        rq  ??(MQ_Q_NAME_LENGTH+1    ??) ; // ReplyToQ model
       char        rqn ??(MQ_Q_NAME_LENGTH+1    ??) ; // ReplyToQ name
       char        rqm ??(MQ_Q_MGR_NAME_LENGTH+1??) ; // ReplyToQ QM
       char        uid ??(MQ_USER_ID_LENGTH+1   ??) ; // Userid of issuer
       char        com ??(MAXCOMMLEN+1          ??) ; // Command to issue

       FULLWORD    timeout                          ; // T timeout
       FULLWORD    verbose                          ; // V flag 0=- 1=+
       FULLWORD    create                           ; // C flag 0=- 1=+
       FULLWORD    mode                             ; // A flag 0== 1=+ 2=-
       char        incommand                        ; // Scanning flag

       FULLWORD parms::getnumb( char * const ptr ) ;    //Private func to get nums out of parms

   public:
                 parms       () { memset(cqm,0,sizeof(cqm))                ; //Constructor
                                  memset(cq ,0,sizeof(cq ))                ; // defaults
                                  memset(rq ,0,sizeof(rq ))                ;
                                  memset(rqn,0,sizeof(rqn))                ;
                                  memset(rqm,0,sizeof(rqm))                ;
                                  memset(uid,0,sizeof(uid))                ;
                                  memset(com,0,sizeof(com))                ;
                                  strcpy(cq ,"SYSTEM.ADMIN.COMMAND.QUEUE") ;
                                  strcpy(rq ,"SYSTEM.MQSC.REPLY.QUEUE"   ) ;
                                  strcpy(rqn,"MQOCOMM.*"                 ) ;
                                  timeout   =  5000                        ;
                                  verbose   =  0                           ;
                                  create    =  1                           ;
                                  mode      =  0                           ;
                                  incommand = 'N'                          ;
                                }


       char    * getcqmname   (char * const dest)
                              {
                                strcpy(dest,cqm) ;
                                return(dest)     ;
                              }

       char    * getcqname    (char * const dest)
                              {
                                strcpy(dest,cq)  ;
                                return(dest)     ;
                              }

       char    * getrqmodel   (char * const dest)
                              {
                                strcpy(dest,rq)  ;
                                return(dest)     ;
                              }

       char    * getrqname    (char * const dest)
                              {
                                strcpy(dest,rqn) ;
                                return(dest)     ;
                              }

       char    * getrqmgr     (char * const dest)
                              {
                                if ( rqm??(0??) != 0x00 )      strcpy(dest,rqm) ;
                                                          else strcpy(dest,cqm) ;
                                return(dest)     ;
                              }

       char    * getuserid    (char * const dest)
                              {
                                strcpy(dest,uid) ;
                                return(dest)     ;
                              }

       char    * getcomm      (char * const dest)
                              {
                                switch ( strlen(this->com) % 4 ) //Pad to multiple of 4 bytes
                                 {
                                   case 1  : strcat(this->com," ") ;
                                   case 2  : strcat(this->com," ") ;
                                   case 3  : strcat(this->com," ") ;
                                   default : break                 ;
                                 }
                                strcpy(dest,com) ; //Return Padded Command
                                return(dest)     ;
                              }

       void      gettimeout   (MQLONG * const to)
                              {
                                *to = timeout    ;
                              }

       void      getcreate    (MQLONG * const to)
                              {
                                *to = create     ;
                              }

       void      getmode      (MQLONG * const to)
                              {
                                *to = mode       ;
                              }

       FULLWORD  isverbose   () { if ( verbose == 1 ) return(TRUE)  ;   //Access func to the pa
                                       else           return(FALSE) ; } ;

       void     parmset ( int     const argc,                       //Function to set
                          char ** const argv,                       // up parms in the
                          char ** const envp  ) ;                   // class

} ;

//
// The parmset function is the part of the parms class
//     that builds the parameters
//
// Valid parms are :
//
//            /Maaaaa     -> The Queue Manager Name
//            /Qaaaaa     -> The Queue Name
//            /Paaaaa     -> The ReplyToQ model
//            /Raaaaa     -> The ReplyToQ name
//            /Naaaaa     -> The ReplyToQ manager
//            /Uaaaaa     -> The Userid
//            /C + -      -> ReplyToQ creation  (Y,N)
//            /A = + -    -> Method of use
//            /Tnnnn      -> Timeout
//            aaaaaaaa    -> The Command
//
void parms::parmset ( int     const argc,
                      char ** const argv,
                      char ** const envp  )
 {
    int     i = 0   ;
    char  * p = 0   ;
    char    c = ' ' ;
    char    s = ' ' ;
    char    v = ' ' ;
    char    d = ' ' ;
    char  * q = 0   ;
    char  * z = 0   ;

    for ( i=1 ; i < argc ; i++ )     // loop through all the parms
        {
          p = argv??(i??) ;          // start of current parm
          z = p           ;          // start of data-type entry
          c = *p          ;          // should be the selector, so ignore it
          s = *(++p)      ;          // the parm key
          p++             ;          // start of parm data
          v = *p          ;          // first of the parm data

          switch ( c )    //See if its a /- or real data
           {
            case '/' :
            case '-' :
                        switch ( s ) //Parm selected
                         {
                           case    'M'  :      // Queue Manager
                           case    'm'  :
                                           if ( incommand != 'Y' ) //Still in parm scan
                                              {
                                                strncpy(this->cqm,p,MQ_Q_MGR_NAME_LENGTH) ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;
                                           break                                ;

                           case    'Q'  :      // Queue - Command Queue
                           case    'q'  :
                                           if ( incommand != 'Y' ) //Still in parm scan
                                              {
                                                strncpy(this->cq,p,MQ_Q_NAME_LENGTH) ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;
                                           break                                ;

                           case    'P'  :      // Queue - Model for ReplyToQ
                           case    'p'  :
                                           if ( incommand != 'Y' ) //Still in parm scan
                                              {
                                                strncpy(this->rq,p,MQ_Q_NAME_LENGTH) ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;
                                           break                                ;

                           case    'R'  :      // Queue - name for ReplyToQ
                           case    'r'  :
                                           if ( incommand != 'Y' ) //Still in parm scan
                                              {
                                                strncpy(this->rqn,p,MQ_Q_NAME_LENGTH) ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;
                                           break                                ;

                           case    'N'  :      // ReplyToQ Queue Manager (if not given, defaults to /M in access func)
                           case    'n'  :
                                           if ( incommand != 'Y' ) //Still in parm scan
                                              {
                                                strncpy(this->rqm,p,MQ_Q_MGR_NAME_LENGTH) ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;
                                           break                                ;

                           case    'U'  :      // Userid
                           case    'u'  :
                                           if ( incommand != 'Y' ) //Still in parm scan
                                              {
                                                strncpy(this->uid,p,MQ_USER_ID_LENGTH) ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;
                                           break                                ;

                           case    't'  :      // Timeout Interval
                           case    'T'  :

                                           timeout = getnumb(p) * 1000 ;
                                           break                       ;

                           case    'c'  :      // ReplyToQ creation
                           case    'C'  :
                                           if ( incommand != 'Y' ) switch ( v ) //Still in parm scann
                                              {
                                                case '+'  :     //Yes
                                                            create  = 1 ;
                                                            break       ;

                                                case '-'  :     //No
                                                            create  = 0 ;
                                                            break       ;

                                                default   : break       ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;

                                           break ;

                           case    'a'  :      // Method of use
                           case    'A'  :
                                           if ( incommand != 'Y' ) switch ( v ) //Still in parm scann
                                              {
                                                case '='  :     //Send & Receive
                                                            mode    = 0 ;
                                                            break       ;

                                                case '+'  :     //Send
                                                            mode    = 1 ;
                                                            break       ;

                                                case '-'  :     //Receive
                                                            mode    = 2 ;
                                                            break       ;

                                                default   : break       ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;

                                           break ;

                           case    'v'  :      // Message usage
                           case    'V'  :
                                           if ( incommand != 'Y' ) switch ( v ) //Still in parm scann
                                              {
                                                case '+'  :     //Yes
                                                            verbose = 1 ;
                                                            break       ;

                                                case '-'  :     //No
                                                            verbose = 0 ;
                                                            break       ;

                                                default   : break       ;
                                              }
                                               else { //In Command logic
                                                       strcat(this->com,z  ) ; //Put back the space
                                                       strcat(this->com," ") ; //Add the parm
                                                    } ;

                                           break ;

                           default      :      // Command
                                           incommand = 'Y'       ; //Ignore / from now on
                                           strcat(this->com,z  ) ; //Put back the space
                                           strcat(this->com," ") ; //Add the parm
                                           break                 ;

                         } ;  //End of parm selection

                        break ; //End of -/ selected

            default: //Non-parm selected

                     incommand = 'Y'       ; //Ignore / from now on
                     strcat(this->com,z  ) ; //Put back the space
                     strcat(this->com," ") ; //Add the parm
                     break                 ;

           } ; //End of 1st char selection

        } ; //End of parm scan loop

    return ;

 }

//
// The getnumb private function within the parms class
//     merely extracts the numeric part of a /Pnnnn
//
FULLWORD parms::getnumb( char * const ptr )
 {
    FULLWORD   numb = 0   ;
    char     * p    = 0   ;
    char       c    = ' ' ;

    for ( p = ptr ; *p != '\0' ; p++ )
        {
          c = *p ;
          if ( c == ' ' ) break ;
          if ( ( c >= '0' ) && ( c <= '9' ) )
                numb = ( numb * 10 ) + ( c - '0' ) ;
        }
    return ( numb ) ;
  }



//
// RAHMQO Class Library
//

class rahmqo
 {
   public :

    char        qm  ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
    char        q   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name
    MQHCONN     qmh                              ; //QM handle
    MQHOBJ      qh                               ; //Q  handle
    MQLONG      rc                               ; //Return Code
    MQLONG      ac                               ; //Reason Code
    MQLONG      lastrc                           ; //Prev Return Code
    MQLONG      lastac                           ; //Prev Reason Code


    rahmqo ( void )               ;  //Contructor for new  use of QM
    rahmqo ( MQHCONN const uqmh)  ;  //Contructor for same use of QM

    MQLONG      conn ( void )               ;  //Connect    to QM
    MQLONG      conn ( char * const uqm )   ;  //  with QM name supplied

    MQLONG      disc ( void )               ;  //Disconnect from QM

    void        cleanup ( void )            ;  //Close and Disconnect

    MQOD        odopen                      ;  //OD for MQOPEN usage
    MQLONG      open (MQLONG const uopts )  ;  //MQOPEN call
    MQLONG      open (char * const uq    ,     //  with Q name supplied
                      MQLONG const uopts )  ;  //
    MQLONG      open (MQOD * const uod   ,     //  with OD and options
                      MQLONG const uopts )  ;  //

    MQLONG      close (MQLONG const uopts ) ;  //MQCLOSE call

    MQLONG      commit   ( void ) ; //Syncpoint UOW
    MQLONG      rollback ( void ) ; //Backout   UOW

    MQLONG      inq ( MQLONG   const  uselnum ,             //MQINQ call
                      MQLONG          usels   ,
                      MQLONG   const  uinum   ,
                      MQLONG *        uiattr  ,
                      MQLONG   const  ucharl  ,
                      char   *        ucattr  ) ;

    MQLONG      set ( MQLONG   const uselnum  ,             //MQSET call
                      MQLONG         usels    ,
                      MQLONG   const uinum    ,
                      MQLONG *       uiattr   ,
                      MQLONG   const ucharl   ,
                      char   *       ucattr   ) ;

    MQMD        pmd                             ;           //PUT Message Dwescriptor
    MQLONG      put ( MQPMO  * const  uopts   ,             //MQPUT call
                      MQLONG   const  bufferl ,
                      char   * const  buffer  ) ;

    MQLONG      put ( MQMD   *        upmd    ,
                      MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  ) ;

                                                            //PUT Zero-Length record
    MQLONG      putz ( MQPMO  * const  uopts     ) ;        //No Tag, Use Instance MD

    MQLONG      putz ( MQMD   *        upmd      ,          //No Tag, use supplied MD
                       MQPMO  * const  uopts     ) ;

    MQLONG      putz ( MQPMO    * const  uopts   ,          //Tag, Use Instance MD
                       MQBYTE24 * const  utag    ) ;

    MQLONG      putz ( MQMD     *        upmd    ,          //Tag, use supplied MD
                       MQPMO    * const  uopts   ,
                       MQBYTE24 * const  utag    ) ;

    MQMD        bmd                                  ;      //Browse Message Descriptor
    MQGMO       bmo                                  ;      //Browse MQGET options
    MQLONG      browse ( MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  ) ;

    MQLONG      browse ( MQMD   *        ubmd      ,
                         MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  ) ;


    MQMD        gmd                               ;         //GET Message Descriptor
    MQLONG      get ( MQGMO  * const  uopts     ,           //MQGET call
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  ) ;

    MQLONG      get ( MQMD   *        ugmd      ,
                      MQGMO  * const  uopts     ,
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  ) ;


    double      gettimeconn     () { return (time_conn     ) ; }
    double      gettimedisc     () { return (time_disc     ) ; }
    double      gettimeopen     () { return (time_open     ) ; }
    double      gettimeclose    () { return (time_close    ) ; }
    double      gettimeinq      () { return (time_inq      ) ; }
    double      gettimeset      () { return (time_set      ) ; }
    double      gettimeput      () { return (time_put      ) ; }
    double      gettimeget      () { return (time_get      ) ; }
    double      gettimecommit   () { return (time_commit   ) ; }
    double      gettimerollback () { return (time_rollback ) ; }
    double      getcputconn     () { return (cput_conn     / CLOCKS_PER_SEC) ; }
    double      getcputdisc     () { return (cput_disc     / CLOCKS_PER_SEC) ; }
    double      getcputopen     () { return (cput_open     / CLOCKS_PER_SEC) ; }
    double      getcputclose    () { return (cput_close    / CLOCKS_PER_SEC) ; }
    double      getcputinq      () { return (cput_inq      / CLOCKS_PER_SEC) ; }
    double      getcputset      () { return (cput_set      / CLOCKS_PER_SEC) ; }
    double      getcputput      () { return (cput_put      / CLOCKS_PER_SEC) ; }
    double      getcputget      () { return (cput_get      / CLOCKS_PER_SEC) ; }
    double      getcputcommit   () { return (cput_commit   / CLOCKS_PER_SEC) ; }
    double      getcputrollback () { return (cput_rollback / CLOCKS_PER_SEC) ; }

 private :

    double      time_conn      ;
    double      time_disc      ;
    double      time_open      ;
    double      time_close     ;
    double      time_inq       ;
    double      time_set       ;
    double      time_put       ;
    double      time_get       ;
    double      time_commit    ;
    double      time_rollback  ;


    double      cput_conn      ;
    double      cput_disc      ;
    double      cput_open      ;
    double      cput_close     ;
    double      cput_inq       ;
    double      cput_set       ;
    double      cput_put       ;
    double      cput_get       ;
    double      cput_commit    ;
    double      cput_rollback  ;
 } ;

//
// rahmqo member function definitions
//

//
// Constructor functions
//
//     without a parm, then this is a new use of a QM
//     with          , then this is an additional access on a
//                     previously connected QM
//

rahmqo::rahmqo ( void )
{
  memset(qm,0,sizeof(qm)) ;
  memset(q ,0,sizeof(q )) ;
  qmh    = 0              ;
  qh     = 0              ;
  rc     = 0              ;
  ac     = 0              ;
  lastrc = 0              ;
  lastac = 0              ;
  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  memset(&pmd,0,sizeof(pmd))                                   ;
  memset(&bmd,0,sizeof(bmd))                                   ;
  memset(&gmd,0,sizeof(gmd))                                   ;

  cput_conn      = 0 ;
  cput_disc      = 0 ;
  cput_open      = 0 ;
  cput_close     = 0 ;
  cput_inq       = 0 ;
  cput_set       = 0 ;
  cput_put       = 0 ;
  cput_get       = 0 ;
  cput_commit    = 0 ;
  cput_rollback  = 0 ;
  time_conn      = 0 ;
  time_disc      = 0 ;
  time_open      = 0 ;
  time_close     = 0 ;
  time_inq       = 0 ;
  time_set       = 0 ;
  time_put       = 0 ;
  time_get       = 0 ;
  time_commit    = 0 ;
  time_rollback  = 0 ;

  return             ;
}

rahmqo::rahmqo ( MQHCONN const uqmh )
{
  memset(qm,0,sizeof(qm)) ;
  memset(q ,0,sizeof(q )) ;
  qmh    = uqmh           ;
  qh     = 0              ;
  rc     = 0              ;
  ac     = 0              ;
  lastrc = 0              ;
  lastac = 0              ;
  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  memset(&pmd,0,sizeof(pmd))                                   ;
  memset(&bmd,0,sizeof(bmd))                                   ;
  memset(&gmd,0,sizeof(gmd))                                   ;

  cput_conn      = 0 ;
  cput_disc      = 0 ;
  cput_open      = 0 ;
  cput_close     = 0 ;
  cput_inq       = 0 ;
  cput_set       = 0 ;
  cput_put       = 0 ;
  cput_get       = 0 ;
  cput_commit    = 0 ;
  cput_rollback  = 0 ;
  time_conn      = 0 ;
  time_disc      = 0 ;
  time_open      = 0 ;
  time_close     = 0 ;
  time_inq       = 0 ;
  time_set       = 0 ;
  time_put       = 0 ;
  time_get       = 0 ;
  time_commit    = 0 ;
  time_rollback  = 0 ;

  return             ;
}

//
// rahmqo destructor functions
//
//   these are not provided
//
//    (Note that a disc call removes access to ALL
//     instances of RAHMQO sharing the Handle. as I
//     am not aiming at a 'proper' C++ implementation
//     I have not provided code to cope with this event).
//

//
// rahmqo functions to access a Queue Manager
//

//
// conn  will connect to a Queue Manager
//       saving the QMH
//
//       The version with no parms will use the
//           pre-set Queue Manager name
//
//       The alternative access is to quote a
//           Queue Manager Name which is the
//           one to be used
//

MQLONG  rahmqo::conn ( char * const uqm )
{
  double c1, c2, t1, t2         ;
  strncpy(qm,uqm,sizeof(qm)-1)  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCONN ( qm, &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_conn += 1.0000 * ( c2 - c1 )      ;
  time_conn += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::conn ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCONN ( qm, &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_conn += 1.0000 * ( c2 - c1 )      ;
  time_conn += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


//
// disc  will disconnect from a Queue Manager
//       deleting the QMH
//


MQLONG  rahmqo::disc ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQDISC ( &qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_disc += 1.0000 * ( c2 - c1 )      ;
  time_disc += 1.0000 * difftime(t2,t1)  ;

  qmh = 0     ;

  return (ac) ;
}

//
// cleanup will Close the Q (if known) and Disconnect from
//         the QM (if known). This will effect other instances
//         which were created with a QM handle
//

void    rahmqo::cleanup ( void )
{
  double c1, c2, t1, t2         ;

  if ( qh  != 0 ) close(MQCO_NONE) ;
  if ( qmh != 0 ) disc ()          ;
  return                           ;
}

//
// commit will Syncpoint ALL accesses to the
//       QM on the given QMH (this will effect
//       other instances of the Class if the
//       QMH is being shared) in the current UOW
//

MQLONG  rahmqo::commit ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCMIT ( qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_commit += 1.0000 * ( c2 - c1 )      ;
  time_commit += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// rollback will backout ALL accesses to the
//       QM on the given QMH (this will effect
//       other instances of the Class if the
//       QMH is being shared) in the current UOW
//

MQLONG  rahmqo::rollback ( void )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQBACK ( qmh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_rollback += 1.0000 * ( c2 - c1 )      ;
  time_rollback += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// open  will provide access to a Queue
//
//       The OPEN options are always supplied
//           by the caller.
//
//       If the version without a Queue Name is
//          used, then the Name from the Instance
//          is taken.
//

MQLONG  rahmqo::open ( MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  strcpy(odopen.ObjectName,q)                                  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqo::open ( char * const uq    ,
                       MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  strncpy(q,uq,sizeof(q)-1)   ;

  memset(&odopen,0,sizeof(odopen))                             ;
  strncpy(odopen.StrucId,MQOD_STRUC_ID,sizeof(odopen.StrucId)) ;
  odopen.Version    = MQOD_VERSION_1                           ;
  odopen.ObjectType = MQOT_Q                                   ;
  strcpy(odopen.ObjectName,q)                                  ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqo::open ( MQOD * const uod   ,
                       MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  memcpy(&odopen,uod,sizeof(odopen)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQOPEN ( qmh, &odopen, uopts, &qh, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_open += 1.0000 * ( c2 - c1 )      ;
  time_open += 1.0000 * difftime(t2,t1)  ;

  memcpy(uod,&odopen,sizeof(odopen))     ;
  return (ac) ;
}

//
// close will stop access to a Queue
//
//       The CLOSE options are always supplied
//           by the caller.
//

MQLONG  rahmqo::close ( MQLONG const uopts )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQCLOSE ( qmh, &qh, uopts, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_close += 1.0000 * ( c2 - c1 )      ;
  time_close += 1.0000 * difftime(t2,t1)  ;

  qh = 0      ;

  return (ac) ;
}

//
// inq   will inquire upon the Attributes of
//       a Queue etc.
//
//       Note that ALL the Selectors stuff
//       has to be supplied by the caller
//

MQLONG  rahmqo::inq ( MQLONG   const  uselnum ,
                      MQLONG          usels   ,
                      MQLONG   const  uinum   ,
                      MQLONG *        uiattr  ,
                      MQLONG   const  ucharl  ,
                      char   *        ucattr  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQINQ ( qmh, qh, uselnum, &usels ,
                   uinum  , uiattr ,
                   ucharl , ucattr ,
                                     &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_inq += 1.0000 * ( c2 - c1 )      ;
  time_inq += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}
//
// set   will set the Attributes of a Queue etc.
//
//       Note that ALL the Selectors stuff
//       has to be supplied by the caller
//

MQLONG  rahmqo::set ( MQLONG   const   uselnum ,
                      MQLONG           usels   ,
                      MQLONG   const   uinum   ,
                      MQLONG *         uiattr  ,
                      MQLONG   const   ucharl  ,
                      char   *         ucattr  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQSET ( qmh, qh, uselnum , &usels  ,
                   uinum   , uiattr  ,
                   ucharl  , ucattr  ,
                                       &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_set += 1.0000 * ( c2 - c1 )      ;
  time_set += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


//
// put       will write a Message into a Queue
//
//           The Message descriptor may either
//           be supplied or the instance version
//           used.
//

MQLONG  rahmqo::put ( MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::put ( MQMD   *        upmd    ,
                      MQPMO  * const  uopts   ,
                      MQLONG   const  bufferl ,
                      char   * const  buffer  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, bufferl, buffer, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// putz      will write a Zero-Length Message into a Queue
//
//           If a Tag is supplied, then this is used as the Message Descriptor
//              this must be an area of MQBYTE24 (24 bytes long)
//
//           The Message descriptor may either
//           be supplied or the instance version
//           used.
//

MQLONG  rahmqo::putz ( MQPMO  * const  uopts )               //No Tag, Use Instance MD
{
  double c1, c2, t1, t2         ;

  memset(&this->pmd.MsgId,0,sizeof(this->pmd.MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::putz ( MQMD   *        upmd    ,              //No Tag, use supplied MD
                       MQPMO  * const  uopts   )
{
  double c1, c2, t1, t2         ;

  memset(upmd->MsgId,0,sizeof(upmd->MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

MQLONG  rahmqo::putz ( MQPMO    * const  uopts   ,            //Tag, Use Instance MD
                       MQBYTE24 * const  utag    )
{
  double c1, c2, t1, t2         ;

  memcpy(&this->pmd.MsgId,utag,sizeof(this->pmd.MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, &pmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::putz ( MQMD     *        upmd    ,            //Tag, use supplied MD
                       MQPMO    * const  uopts   ,
                       MQBYTE24 * const  utag    )
{
  double c1, c2, t1, t2         ;

  memcpy(upmd->MsgId,utag,sizeof(upmd->MsgId)) ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQPUT ( qmh, qh, upmd, uopts, 0, 0, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_put += 1.0000 * ( c2 - c1 )      ;
  time_put += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
//    Browse    will non-destructively read a Queue
//
//              Versions are supplied that use an
//              instance Message dewscriptor, or
//              one supplied by the caller
//

MQLONG  rahmqo::browse ( MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  memset(&bmo,0,sizeof(bmo))                                 ;
  memcpy(&bmo.StrucId,MQGMO_STRUC_ID,sizeof(bmo.StrucId))    ;
  bmo.Version        = MQGMO_VERSION_1                       ;
  bmo.Options        = MQGMO_NO_WAIT+MQGMO_BROWSE_NEXT+MQGMO_ACCEPT_TRUNCATED_MSG+
                       MQGMO_FAIL_IF_QUIESCING               ;
  bmo.WaitInterval   = MQWI_UNLIMITED                        ;

  memset(&bmd,0,sizeof(bmd))                                 ;
  memcpy(&bmd.StrucId,MQMD_STRUC_ID,sizeof(bmd.StrucId))     ;
  bmd.Version        = MQMD_VERSION_1                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, &bmd, &bmo, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::browse ( MQMD   *        ubmd      ,
                         MQLONG   const  bufferl   ,
                         char   * const  buffer    ,
                         MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  memset(&bmo,0,sizeof(bmo))                                 ;
  memcpy(&bmo.StrucId,MQGMO_STRUC_ID,sizeof(bmo.StrucId))    ;
  bmo.Version        = MQGMO_VERSION_1                       ;
  bmo.Options        = MQGMO_NO_WAIT+MQGMO_BROWSE_NEXT+MQGMO_ACCEPT_TRUNCATED_MSG+
                       MQGMO_CONVERT+MQGMO_FAIL_IF_QUIESCING ;
  bmo.WaitInterval   = MQWI_UNLIMITED                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, ubmd, &bmo, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::get ( MQGMO  * const  uopts     ,           //MQGET call
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  memset(&gmd,0,sizeof(gmd))                                 ;
  memcpy(&gmd.StrucId,MQMD_STRUC_ID,sizeof(gmd.StrucId))     ;
  bmo.Version        = MQMD_VERSION_1                        ;

  memset(&gmd,0,sizeof(gmd))                                 ;
  memcpy(&gmd.StrucId,MQMD_STRUC_ID,sizeof(gmd.StrucId))     ;
  gmd.Version        = MQMD_VERSION_1                        ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, &gmd, uopts, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}


MQLONG  rahmqo::get ( MQMD   *        ugmd      ,
                      MQGMO  * const  uopts     ,
                      MQLONG   const  bufferl   ,
                      char   * const  buffer    ,
                      MQLONG *        udatalen  )
{
  double c1, c2, t1, t2         ;

  lastrc = rc ;
  lastac = ac ;

  c1 = clock()    ;
  t1 = time(NULL) ;
  MQGET ( qmh, qh, ugmd, uopts, bufferl, buffer, udatalen, &rc, &ac ) ;
  c2 = clock()    ;
  t2 = time(NULL) ;

  cput_get += 1.0000 * ( c2 - c1 )      ;
  time_get += 1.0000 * difftime(t2,t1)  ;

  return (ac) ;
}

//
// End of RAHMQO Class definition and functions
//


//
// Main procedure - Issue a Command to the System Command Queue
//
int  main ( int     const argc,
            char ** const argv,
            char ** const envp  )
{
 parms       myparms        ;
 rahmqo      cmq            ; //MQ instances - command
 rahmqo      rmq            ; //             - reply

 char        cqm  ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name
 char        cq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - command
 char        rq   ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - replyToq model
 char        rqn  ??(MQ_Q_NAME_LENGTH+1    ??) ; //Q  name - replyToq name
 char        rqm  ??(MQ_Q_MGR_NAME_LENGTH+1??) ; //QM name - replyToq manager
 char        uid  ??(MQ_USER_ID_LENGTH+1   ??) ; //Userid of issuer
 MQLONG      to   = 5000                       ; //Timeout
 MQLONG      cr   = 1                          ; //ReplyToQ creation
 MQLONG      mo   = 0                          ; //Send/Receive/Both mode 1 2 0

 char        command  ??(MAXCOMMLEN+5??)       ; //Command (rounded to 4 bytes!)

 char        pcfarea  ??(MQCFH_STRUC_LENGTH        +    //PCF area - header
                         MQCFIN_STRUC_LENGTH       +    //         - integer parm
                         MQCFST_STRUC_LENGTH_FIXED +    //         - string parm
                         MAXCOMMLEN+5             ??) ; //         - command

 char      * ptrpcf  = (char   *) pcfarea                            ; //Set PCF area pointers
 MQCFH     * ptrpcf1 = (MQCFH  *) pcfarea                            ; // within PCF area
 MQCFIN    * ptrpcf2 = (MQCFIN *)&pcfarea??(MQCFH_STRUC_LENGTH??)    ;
 MQCFST    * ptrpcf3 = (MQCFST *)&pcfarea??(MQCFH_STRUC_LENGTH+MQCFIN_STRUC_LENGTH??)                           ;
 char      * ptrpcf4 = (char   *)&pcfarea??(MQCFH_STRUC_LENGTH+MQCFIN_STRUC_LENGTH+MQCFST_STRUC_LENGTH_FIXED??) ;

 MQLONG      pcflen   = 0    ; //Length of PCF area
 char     *  buffer   = 0    ; //Pointer to ReplyToQ buffer

 MQCFH     * bufpcf1         ; //Declare pointers to PCF
 MQCFIN    * bufpcf2i        ; // areas returned within
 MQCFST    * bufpcf2s        ; // the ReplyToQ buffer
 MQCFIL    * bufpcf2il       ;
 MQCFSL    * bufpcf2sl       ;

 MQLONG      gopts    = 0    ; //MQ Control Blocks
 MQLONG      popts    = 0    ;
 MQGMO       gmo             ;
 MQPMO       pmo             ;
 MQOD        od              ;

 MQLONG      mqrc     = 0    ; //MQ Return codes
 MQLONG      mqac     = 0    ;
 MQLONG      crc      = 0    ;
 MQLONG      rrc      = 0    ;
 MQLONG      ac       = 0    ;
 MQLONG      rc       = 0    ;

 FULLWORD    recno   =  0    ; //Number of ReplyToQ Message

 myparms.parmset(argc, argv, envp) ;  //set the parms

 myparms.getcqmname(cqm)           ;  //Get the data
 myparms.getcqname (cq )           ;  // from the
 myparms.getrqmodel(rq )           ;  // parms
 myparms.getrqname (rqn)           ;
 myparms.getrqmgr  (rqm)           ;
 myparms.getuserid (uid)           ;
 myparms.getcomm   (command)       ;
 myparms.gettimeout(&to)           ;
 myparms.getcreate (&cr)           ;
 myparms.getmode   (&mo)           ;

 //Display program parms

 cout << "!Program parms are: " ;
 for ( int i=0 ; i<argc ; cout << argv??(i++??) << " " ) ;
 cout << endl ;

 if ( argc == 1 )                    //Show parms ONLY if none given
    {
      cout << "\nMQOCOMM will issue a Command to the Queue Manager\n"
           << "Parms for MQOCOMM are /M<source queue manager>  /Q<command input Queue>\n"
           << "                      /P<Model for ReplyToQ>    /R<ReplyToQ name>\n"
           << "                      /N<ReplyToQ Manager>      /U<Userid of issuer>\n"
           << "                      /T<timeout>               /C<ReplyToQ creation>\n"
           << "                      /A<mode>                  /V<verbose>\n"
           << "                      <command>\n"
           << " <command> must come after all the parms\n\n"
           << " The default command queue  (/Q) is SYSTEM.ADMIN.COMMAND.QUEUE\n"
           << " The default ReplyToQ model (/P) is SYSTEM.MQSC.REPLY.QUEUE\n"
           << " The default ReplyToQ name  (/R) is MQOCOMM.*\n"
           << " The default ReplyToQ mgr   (/N) is /M\n"
           << " Timeout defaults to 5 seconds\n"
           << " /C+ Create ReplyToQ (/P /R) (default)  /C- use extant ReplyToQ (/R)\n"
           << " /A= Send Command and Receive results (default)\n"
           << " /A+ Send only to /R                    /A- Receive only from /R\n"
           << " /V+ display                            /V- no messages (default)\n\n"
           << " The Command Server (AMQPCSEA) must be running\n\n"
           ;
      return (0) ;
    }

 //Let on which operational mode is active

 if ( myparms.isverbose() ) switch ( mo )
    {
      case  2 :  cout << "!Receiving results only"   << endl ;
                 break ;

      case  1 :  cout << "!Sending command only"     << endl ;
                 break ;

      case  0 :
      default :  cout << "!Processing whole command" << endl ;
                 break ;
    }

 //Display the command

 cout << "!Command is <" << command << ">" ;
 switch ( mo )
    {
      case  2 :  // /A- Receive only
                 cout << " which is ignored due to /A- " << endl ;
                 break  ;

      case  1 :  // /A+ Send only
      case  0 :  // /A= Send & Receive
      default :
                 cout << endl ;
                 break ;
    }

 //In Both Mode, the QM and ReplyToQM have to be the same, so force it so to be

 switch ( mo )
    {
      case  2 :  // /A- Receive only
      case  1 :  // /A+ Send only
                 break ;

      case  0 :  // /A= Send & Receive
      default :
                 if ( strcmp(cqm,rqm) != 0 )
                    {
                      if ( myparms.isverbose() )
                           cout << "!In Both Mode, the QMs must be the same, so"
                                << " /N forced to /M due to /A= " << endl ;
                      memcpy(rqm, cqm, sizeof(rqm)) ;
                    }
                 break ;
    }

 //Force /C- (reuse extant ReplyToQ) if doing /A- (could create with /A+)

 switch ( mo )
    {
      case  2 :  // /A- Receive only
                 if ( cr != 0 )
                    {
                      cr = 0 ;
                      if ( myparms.isverbose() )
                           cout << "!/C- forced due to /A- " << endl ;
                    }
                 break  ;

      case  1 :  // /A+ Send only
      case  0 :  // /A= Send & Receive
      default :
                 break ;
    }

 //Force /C- (reuse extant ReplyToQ) if the QMs are different
 //          (as cannot create it on different QM)
 //
 //               Note the above /A= forcing of same QMs has to be
 //                    before this test (or else /A= disappears!)

 switch ( cr )
    {
      case  0 :  // /C- Reuse extant ReplyToQ
                 break ;

      case  1 :  // /C+ Create a ReplyToQ
      default :
                 if ( strcmp(cqm, rqm) != 0 )
                    {
                      cr = 0 ;
                      if ( myparms.isverbose() )
                           cout << "!/C- forced due to different QMs " << endl ;
                    }
                 break  ;
    }

 //Set the Put options (for the Command) and Get options (for the reply)

 gopts  = MQGMO_NO_SYNCPOINT         + //Access is outside UoW
          MQGMO_WAIT                 +
          MQGMO_ACCEPT_TRUNCATED_MSG +
          MQGMO_FAIL_IF_QUIESCING    ;

 popts  = MQPMO_NO_SYNCPOINT         +
          MQPMO_DEFAULT_CONTEXT      + //Needed to bluff PCF Security!
          MQPMO_FAIL_IF_QUIESCING    ;

 //If no command, then nothing to do if in Send or Both mode

 if ( ( mo != 2 ) && ( strlen(command) == 0 ) ) rc = -7 ;

 //Connect to the QM, open the Command Queue, and create the ReplyToQ

 if ( rc == 0 )                      // Connect to QM
    {
      if ( myparms.isverbose() ) cout << "!Connecting to QM " << cqm << " " ;
      ac   = cmq.conn(cqm)   ;
      mqrc = cmq.lastrc      ;
      mqac = cmq.lastrc      ;

      if ( ac == 0 ) rmq.qmh = cmq.qmh ;
      if ( ac != 0 ) rc = ac           ;
      if ( myparms.isverbose() ) cout << "RC=" << rc << endl      ;
    }

if ( rc == 0 )                      // Open the command Q for Put access for Send or both
   {
      switch ( mo )
       {
         case  2 :  // /A- Receive only
                    break ;

         case  1 :  // /A+ Send only
         case  0 :  // /A= Send & Receive
         default :
                    if ( myparms.isverbose() ) cout << "!Opening Command Q " << cq << " for Output " ;
                    ac   = cmq.open(cq,(MQOO_OUTPUT+MQOO_FAIL_IF_QUIESCING))   ;
                    mqrc = cmq.lastrc     ;
                    mqac = cmq.lastrc     ;

                    if ( ac != 0 ) rc = ac ;
                    if ( myparms.isverbose() ) cout << "RC=" << rc << endl ;
                    break ;
       }
   }

 if ( rc == 0 )  //Create or reuse the ReplyToQ
    {
     switch ( cr )
      {
        case 0 :                     // Open the named ReplyToQ for Get access,
                                     //      as the Queue exists, it is named by /S
                                     //
                                     // If in SEND mode, then this call is done to
                                     //    ensure that the ReplyToQ actually exists

                 if ( myparms.isverbose() ) cout << "!Reusing ReplyToQ " << rqn << endl ;
                 memset(&od,0,sizeof(od))                             ;
                 memcpy(od.StrucId,MQOD_STRUC_ID,sizeof(od.StrucId))  ;
                 od.Version        = MQOD_VERSION_1                   ;
                 od.ObjectType     = MQOT_Q                           ;
                 strcpy(od.ObjectName,rqn)                            ;

                 if ( myparms.isverbose() ) cout << "!Opening Object ReplyToQ " << od.ObjectName
                                                 << " for Destructive access " ;

                 ac   = rmq.open(&od,(MQOO_INPUT_EXCLUSIVE
                                           +MQOO_BROWSE+MQOO_SAVE_ALL_CONTEXT
                                           +MQOO_FAIL_IF_QUIESCING))                           ;
                 mqrc = rmq.lastrc     ;
                 mqac = rmq.lastrc     ;

                 if ( ac != 0 ) rc = ac ;
                 if ( myparms.isverbose() ) cout << "RC=" << rc << endl ;

                 break ;

        case 1  :                    // Use a  model to create the named ReplyToQ for Get access,
        default :                    //      as the Queue does not exist, it is named by
                                     //      /S and modeled on /R

                 if ( myparms.isverbose() ) cout << "!Creating ReplyToQ " << rqn << endl ;

                                                // Open the ReplyToQ for Get access
                                                //  as this is a Model, the Dynamic Queue is
                                                //  built, and its name returned in the OD
                 memset(&od,0,sizeof(od))                             ;
                 memcpy(od.StrucId,MQOD_STRUC_ID,sizeof(od.StrucId))  ;
                 od.Version        = MQOD_VERSION_1                   ;
                 od.ObjectType     = MQOT_Q                           ;
                 strcpy(od.ObjectName,rq)                             ;
                 strcpy(od.DynamicQName,rqn)                          ;

                 if ( myparms.isverbose() ) cout << "!Opening Object ReplyToQ " << od.ObjectName
                                                 << " (Dynamic Name "           << od.DynamicQName
                                                 << ") for Destructive access " ;

                 ac   = rmq.open(&od,(MQOO_INPUT_EXCLUSIVE
                                           +MQOO_BROWSE+MQOO_SAVE_ALL_CONTEXT
                                           +MQOO_FAIL_IF_QUIESCING))                           ;
                 mqrc = rmq.lastrc     ;
                 mqac = rmq.lastrc     ;

                 if ( ac != 0 ) rc = ac ;
                 if ( myparms.isverbose() ) cout << "RC=" << rc << endl ;

                 if ( ac == 0 )
                    {
                     memset(&rq,0,sizeof(rq))                        ;
                     strncpy(rq,od.ObjectName,sizeof(od.ObjectName)) ;
                     if ( myparms.isverbose() ) cout << "!The ReplyToQ name is " << rq << endl  ;
                    }
                 break ;

      } // End of the /S select

    } // End of the open Queue Block

 //Obtain a buffer for the ReplyToQ messages

 if ( rc == 0 )                      // Getmain the Buffer (always max file size)
    {                                //         for Both and receive
      switch ( mo )
       {
         case  1 :  // /A+ Send only
                    break ;

         case  2 :  // /A- Receive only
         case  0 :  // /A= Send & Receive
         default :
                   if ( myparms.isverbose() ) cout << "!Getmaining the ReplyToQ Buffer " ;
                   buffer = (char *)malloc(MAXUHALFWORD) ;
                   if ( myparms.isverbose() ) cout << "RC=" << rc << endl   ;
                   if ( buffer == 0 ) rc = -8    ;
                   bufpcf1 = (MQCFH *) buffer    ;
                   break                         ;
       }
    }

 //Build the PCF ESCAPE command

 if ( rc == 0 )                      // Now build the PCF command
    {                                //     for Send and Both only
      switch ( mo )
       {

         case  2 :  // /A- Receive only
                    break ;

         case  1 :  // /A+ Send only
         case  0 :  // /A= Send & Receive
         default :
                    memset(pcfarea,0,sizeof(pcfarea))             ;
                    ptrpcf1->Type           = MQCFT_COMMAND       ; // Header - 2 parms
                    ptrpcf1->StrucLength    = MQCFH_STRUC_LENGTH  ;
                    ptrpcf1->Version        = MQCFH_VERSION_1     ;
                    ptrpcf1->Command        = MQCMD_ESCAPE        ;
                    ptrpcf1->MsgSeqNumber   = 1                   ;
                    ptrpcf1->Control        = MQCFC_LAST          ; // only one command issued
                    ptrpcf1->ParameterCount = 2                   ; // 2 parms - type + data
                    ptrpcf2->Type           = MQCFT_INTEGER       ; // Parm1 is ESCAPE
                    ptrpcf2->StrucLength    = MQCFIN_STRUC_LENGTH ;
                    ptrpcf2->Parameter      = MQIACF_ESCAPE_TYPE  ;
                    ptrpcf2->Value          = MQET_MQSC           ;
                    ptrpcf3->Type           = MQCFT_STRING        ; // Parm2 is ESCAPE data
                    ptrpcf3->StrucLength    = MQCFST_STRUC_LENGTH_FIXED + strlen(command) ;
                    ptrpcf3->Parameter      = MQCACF_ESCAPE_TEXT  ;
                    ptrpcf3->CodedCharSetId = MQCCSI_DEFAULT      ;
                    ptrpcf3->StringLength   = strlen(command)     ;
                    strcpy(ptrpcf4,command)                       ;

                    pcflen =   MQCFH_STRUC_LENGTH
                             + MQCFIN_STRUC_LENGTH
                             + MQCFST_STRUC_LENGTH_FIXED
                             + strlen(command)                    ;
                    break                                         ;

       }
    }

// Now write the Command to the relevant Command Queue

 if ( rc == 0)                       //Now write to the Queue
    {                                //    for Send and Both
      switch ( mo )
       {

         case  2 :  // /A- Receive only
                    break ;

         case  1 :  // /A+ Send only
         case  0 :  // /A= Send & Receive
         default :
                    memset(&pmo,0,sizeof(pmo))  ;               //Set Put Message Options
                    memcpy(&pmo.StrucId,MQPMO_STRUC_ID,sizeof(pmo.StrucId))          ;
                    pmo.Version        = MQPMO_VERSION_1                             ;
                    pmo.Options        = popts                                       ;

                    memset(&cmq.pmd,0,sizeof(cmq.pmd))  ;     //Set Put Message Desc
                    memcpy(&cmq.pmd.StrucId,MQMD_STRUC_ID,sizeof(cmq.pmd.StrucId))     ;
                    memcpy(&cmq.pmd.Format ,MQFMT_ADMIN  ,sizeof(cmq.pmd.Format ))     ;
                    cmq.pmd.Version        = MQMD_VERSION_1                            ;
                    cmq.pmd.Report         = MQRO_NONE                                 ;
                    cmq.pmd.MsgType        = MQMT_REQUEST                              ;
                    cmq.pmd.Persistence    = MQPER_NOT_PERSISTENT                      ;
                    cmq.pmd.Expiry         = MQEI_UNLIMITED                            ;
                    cmq.pmd.Feedback       = MQFB_NONE                                 ;
                    cmq.pmd.Encoding       = MQENC_NATIVE                              ;
                    cmq.pmd.CodedCharSetId = MQCCSI_Q_MGR                              ;
                    cmq.pmd.Priority       = MQPRI_PRIORITY_AS_Q_DEF                   ;
                    memcpy(&cmq.pmd.ReplyToQMgr,rqm,sizeof(cmq.pmd.ReplyToQMgr))       ;
                    memcpy(&cmq.pmd.ReplyToQ,od.ObjectName,sizeof(od.ObjectName))      ;
                    memcpy(&cmq.pmd.UserIdentifier,uid,sizeof(cmq.pmd.UserIdentifier)) ;

                    crc = cmq.put( &pmo, pcflen, (char *)ptrpcf1 )        ;

                    if ( myparms.isverbose() ) cout << "!MQ PUT to Command Queue RC=" << crc << endl ;
                    rc = crc ;
                    break    ;
       }
    }

 //Read the ReplyToQ, and display the contained messages

 if ( rc == 0 )                      // Read Q and Display
    {                                //  for both and Receive
      switch ( mo )
       {
         case  1 :  // /A+ Send only
                    break ;

         case  2 :  // /A- Receive only
         case  0 :  // /A= Send & Receive
         default :
                    for ( recno = 1 ; ; recno++ )
                        {
                          MQLONG   i      = 0   ;
                          char     exiter = 'N' ;
                          char   * p      = 0   ;
                          MQLONG   mrc    = 0   ;
                          MQLONG   frc    = 0   ;
                          FULLWORD reclen = 0   ;

                          bufpcf1   = (MQCFH  *) buffer                         ; //Point to
                          bufpcf2i  = (MQCFIN *)&buffer??(MQCFH_STRUC_LENGTH??) ; //Buffer
                          bufpcf2s  = (MQCFST *)&buffer??(MQCFH_STRUC_LENGTH??) ; //Structures
                          bufpcf2il = (MQCFIL *)&buffer??(MQCFH_STRUC_LENGTH??) ;
                          bufpcf2sl = (MQCFSL *)&buffer??(MQCFH_STRUC_LENGTH??) ;

                          memset(&gmo,0,sizeof(gmo))                               ; //Build the
                          memcpy(&gmo.StrucId,MQGMO_STRUC_ID,sizeof(gmo.StrucId))  ; //GMO, allow
                          gmo.Version        = MQGMO_VERSION_1                     ; //n seconds
                          gmo.Options        = gopts                               ; //for the
                          if ( to != 0 ) gmo.WaitInterval = to                     ; //reply
                               else      gmo.WaitInterval = MQWI_UNLIMITED         ;

                          memset(&rmq.gmd,0,sizeof(rmq.gmd))                              ; //Clear out the
                          memcpy(&rmq.gmd.StrucId,MQMD_STRUC_ID,sizeof(rmq.gmd.StrucId))  ; // MD and the
                          rmq.gmd.Version        = MQMD_VERSION_1                         ; // buffer
                          memset(buffer,0,MAXUHALFWORD)                                   ;
                          reclen   = 0                                                    ;

                          rrc = rmq.get ( &gmo, MAXUHALFWORD, buffer, &reclen )           ; //Obtain ReplyToQ Message

                          switch ( rrc ) //Print obtained message
                           {
                             case MQRC_NONE      :
                                  rc   = 0       ;

                                  if ( myparms.isverbose() )
                                       cout << "!Message "   << recno
                                            << " = MQ Message " << bufpcf1->MsgSeqNumber
                                            << ((bufpcf1->Control == MQCFC_LAST) ? " (last)" : " (more)")
                                            << " has MQ RC "        << bufpcf1->Reason
                                            << ", "                 << bufpcf1->ParameterCount
                                            << " parms"
                                            << endl ;

                                  if ( bufpcf1->Control == MQCFC_LAST) exiter = 'L' ; //Last means just that!
                                                                                      // .. so multiple reply sets
                                                                                      //    require multiple usages
                                                                                      //    of MQOCOMM to retrieve
                                                                                      //    them all

                                  if ( bufpcf1->ParameterCount != 0 ) //Do not attempt to ignore null messages
                                     {                                // when printing the results
                                      MQLONG   parms    ;
                                      int      i        ;
                                      int      j        ;
                                      MQLONG   parmtype ;
                                      MQLONG   parmsize ;
                                      MQLONG   iparm    ;
                                      MQLONG   ival     ;
                                      MQCFIL * ip       ;
                                      MQCFSL * sp       ;

                                      parms = bufpcf1->ParameterCount ;

                                      for (i=0 ; i < parms ; i++ )
                                          {
                                           parmtype = bufpcf2i->Type        ; //All types share a
                                           parmsize = bufpcf2i->StrucLength ; // common prefix
                                           switch ( parmtype )
                                            {
                                              case MQCFT_INTEGER  : //Convert to attr=value
                                                   if ( myparms.isverbose() )
                                                        cout << " Integer Parm " << bufpcf2i->Parameter
                                                             << " Value "        << bufpcf2i->Value
                                                             << endl ;
                                                   break        ;

                                              case MQCFT_INTEGER_LIST : //Convert to attr=value pairs
                                                   if ( ( bufpcf2il->Count != 0 ) && ( myparms.isverbose() ) )
                                                      {
                                                        cout << " Integer parm " << bufpcf2il->Parameter
                                                             << " Values "  ;
                                                        for ( j=0 ; j < bufpcf2il->Count ; j++ )
                                                            {
                                                              ip = bufpcf2il + (sizeof(MQLONG) * j) ;
                                                              cout << ip->Values << " "             ;
                                                            } ;
                                                        cout << endl ;
                                                      } ;
                                                   break        ;

                                              case MQCFT_STRING  : //Print string LEFT Justified
                                                   printf(" %-*s\n",bufpcf2s->StringLength,bufpcf2s->String) ;
                                                   break        ;

                                              case MQCFT_STRING_LIST : //Print Strings LEFT Justified
                                                   if ( bufpcf2sl->Count != 0 )
                                                      {
                                                       for ( j=0 ; j < bufpcf2sl->Count ; j++ )
                                                           {
                                                             sp = bufpcf2sl + ( (bufpcf2sl->StringLength) * j)   ;
                                                             printf(" %-*s\n",bufpcf2sl->StringLength,sp->Strings) ;
                                                           } ;
                                                      } ;
                                                   break        ;

                                              default :
                                                   break ;

                                            } ;

                                           bufpcf2i  = (MQCFIN *)( (char *)bufpcf2i  + parmsize ) ; //Bump
                                           bufpcf2s  = (MQCFST *)( (char *)bufpcf2s  + parmsize ) ; // up
                                           bufpcf2il = (MQCFIL *)( (char *)bufpcf2il + parmsize ) ; // structure
                                           bufpcf2sl = (MQCFSL *)( (char *)bufpcf2sl + parmsize ) ; // pointers

                                          } ; //End of Parameter Printing loop

                                     } ; //End of Parameter Listing

                                  break          ; //End of good message processing

                             default             :
                                  if ( myparms.isverbose() ) cout << "!MQ GET on ReplyToQ error RC=" << rrc
                                                                  <<  " on message "    << recno << endl ;
                                  exiter = 'G'  ;
                                  rc  = rrc     ;
                                  break         ;

                           } //End of ReplyToQ Message processing

                            if ( exiter != 'N' ) break ; //Get next message
                                                         //unless error/EOQ occured

                        } // End of Message Processing FOR loop
                    break ;
       }  // End of Receive/Both Selection block

    } // End of Queue/Queue Processing Block

 if ( rc == 0 ) switch ( mo )
    {
      case  1 :  // /A+ Send only
                 break ;

      case  2 :  // /A- Receive only
      case  0 :  // /A= Send & Receive
      default :
                 cout << "!All ReplyToQ records obtained" << endl ;
                 break ;
    }

// End of processing

//Close all the Queues (even if we have not opened them, as rc ignored!)

 if ( rc == 0 )                      // Close the Q - ReplyToQ
    {                                //  ..this also deletes a TEMPDYN Q
      if ( myparms.isverbose() ) cout << "!Closing the ReplyToQ " ;
      ac   = rmq.close(MQCO_NONE)  ;
      mqrc = rmq.lastrc            ;
      mqac = rmq.lastrc            ;

      if ( ac != 0 ) rc = ac ;
      if ( myparms.isverbose() ) cout << "RC=" << rc << endl ;
    }

 if ( rc == 0 )                      // Close the Q - Command
    {
      if ( myparms.isverbose() ) cout << "!Closing the command Q " ;
      ac   = cmq.close(MQCO_NONE)  ;
      mqrc = cmq.lastrc            ;
      mqac = cmq.lastrc            ;

      if ( ac != 0 ) rc = ac ;
      if ( myparms.isverbose() ) cout << "RC=" << rc << endl ;
    }

if ( rc == 0 )                      // Disconnect from Command QM
    {
      if ( myparms.isverbose() ) cout << "!Disconnecting from Command QM " ;
      ac   = cmq.disc()            ;
      mqrc = cmq.lastrc            ;
      mqac = cmq.lastrc            ;

      if ( ac != 0 ) rc = ac ;
      if ( myparms.isverbose() ) cout << "RC=" << rc << endl ;
    }


 if ( buffer != 0 ) free(buffer)      ; //Freemain the buffer
 rmq.cleanup()                        ; //Stop all accesses
 cmq.cleanup()                        ;

 cout << "!End of MQOCOMM processing " << endl ;

 return(rc) ;                           //Return RC

}

// End of MQOCOMM.CPP source
